@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: cpl_port.h 37456 2017-02-25 18:35:03Z rouault $
+ * $Id: cpl_port.h efdea747194ad7dc810fb30654f4cf3f09756013 2021-04-15 10:09:43 +0100 Andrew C Aitchison $
  *
  * Project:  CPL - Common Portability Library
  * Author:   Frank Warmerdam, warmerdam@pobox.com
@@ -8,7 +8,7 @@
  *
  ******************************************************************************
  * Copyright (c) 1998, 2005, Frank Warmerdam <warmerdam@pobox.com>
- * Copyright (c) 2008-2013, Even Rouault <even dot rouault at mines-paris dot org>
+ * Copyright (c) 2008-2013, Even Rouault <even dot rouault at spatialys.com>
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -123,7 +123,7 @@
 #endif
 
 /* Needed for std=c11 on Solaris to have strcasecmp() */
-#if defined(GDAL_COMPILATION) && defined(__sun__) && __STDC_VERSION__ >= 201112L && _XOPEN_SOURCE < 600
+#if defined(GDAL_COMPILATION) && defined(__sun__) && (__STDC_VERSION__ + 0) >= 201112L && (_XOPEN_SOURCE + 0) < 600
 #ifdef _XOPEN_SOURCE
 #undef _XOPEN_SOURCE
 #endif
@@ -179,11 +179,19 @@
 /*      Which versions of C++ are available.                            */
 /* -------------------------------------------------------------------- */
 
-#ifdef __cplusplus
-#  if __cplusplus >= 201103L
-#    define HAVE_CXX11 1
+/* MSVC fails to define a decent value of __cplusplus. Try to target VS2015*/
+/* as a minimum */
+
+#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS)
+#  if !(__cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1900))
+#    error Must have C++11 or newer.
+#  endif
+#  if __cplusplus >= 201402L || (defined(_MSVC_LANG) && _MSVC_LANG >= 201402L)
+#    define HAVE_CXX14 1
+#  endif
+#  if __cplusplus >= 201703L || (defined(_MSVC_LANG) && _MSVC_LANG >= 201703L)
+#    define HAVE_CXX17 1
 #  endif
-/* TODO(schwehr): What are the correct tests for C++ 14 and 17? */
 #endif  /* __cplusplus */
 
 /*---------------------------------------------------------------------
@@ -215,35 +223,25 @@ typedef unsigned char   GByte;
 typedef int             GBool;
 #endif
 
+/*! @cond Doxygen_Suppress */
+#ifdef __cplusplus
+#define CPL_STATIC_CAST(type, expr) static_cast<type>(expr)
+#define CPL_REINTERPRET_CAST(type, expr) reinterpret_cast<type>(expr)
+#else
+#define CPL_STATIC_CAST(type, expr) ((type)(expr))
+#define CPL_REINTERPRET_CAST(type, expr) ((type)(expr))
+#endif
+/*! @endcond */
+
 /* -------------------------------------------------------------------- */
 /*      64bit support                                                   */
 /* -------------------------------------------------------------------- */
 
 #if defined(WIN32) && defined(_MSC_VER)
-
 #define VSI_LARGE_API_SUPPORTED
-typedef __int64          GIntBig;
-typedef unsigned __int64 GUIntBig;
+#endif
 
-/** Minimum GIntBig value */
-#define GINTBIG_MIN     ((GIntBig)(0x80000000) << 32)
-/** Maximum GIntBig value */
-#define GINTBIG_MAX     (((GIntBig)(0x7FFFFFFF) << 32) | 0xFFFFFFFFU)
-/** Maximum GUIntBig value */
-#define GUINTBIG_MAX     (((GUIntBig)(0xFFFFFFFFU) << 32) | 0xFFFFFFFFU)
-
-#define CPL_HAS_GINT64 1
-
-/** Signed 64 bit integer type */
-typedef GIntBig          GInt64;
-/** Unsigned 64 bit integer type */
-typedef GUIntBig         GUInt64;
-
-#define GINT64_MIN      GINTBIG_MIN
-#define GINT64_MAX      GINTBIG_MAX
-#define GUINT64_MAX     GUINTBIG_MAX
-
-#elif HAVE_LONG_LONG
+#if HAVE_LONG_LONG
 
 /** Large signed integer type (generally 64-bit integer type).
  *  Use GInt64 when exactly 64 bit is needed */
@@ -253,11 +251,11 @@ typedef long long        GIntBig;
 typedef unsigned long long GUIntBig;
 
 /** Minimum GIntBig value */
-#define GINTBIG_MIN     ((GIntBig)(0x80000000) << 32)
+#define GINTBIG_MIN     (CPL_STATIC_CAST(GIntBig, 0x80000000) << 32)
 /** Maximum GIntBig value */
-#define GINTBIG_MAX     (((GIntBig)(0x7FFFFFFF) << 32) | 0xFFFFFFFFU)
+#define GINTBIG_MAX     ((CPL_STATIC_CAST(GIntBig, 0x7FFFFFFF) << 32) | 0xFFFFFFFFU)
 /** Maximum GUIntBig value */
-#define GUINTBIG_MAX     (((GUIntBig)(0xFFFFFFFFU) << 32) | 0xFFFFFFFFU)
+#define GUINTBIG_MAX    ((CPL_STATIC_CAST(GUIntBig, 0xFFFFFFFFU) << 32) | 0xFFFFFFFFU)
 
 /*! @cond Doxygen_Suppress */
 #define CPL_HAS_GINT64 1
@@ -279,14 +277,8 @@ typedef GUIntBig         GUInt64;
 
 #else
 
-// NOTE: we don't really support such platforms ! Many things might break
+#error "64bit integer support required"
 
-typedef long             GIntBig;
-typedef unsigned long    GUIntBig;
-
-#define GINTBIG_MIN     INT_MIN
-#define GINTBIG_MAX     INT_MAX
-#define GUINTBIG_MAX     UINT_MAX
 #endif
 
 #if SIZEOF_VOIDP == 8
@@ -299,9 +291,7 @@ typedef int              GPtrDiff_t;
 
 #ifdef GDAL_COMPILATION
 #if HAVE_UINTPTR_T
-#if !defined(_MSC_VER) || _MSC_VER > 1500
 #include <stdint.h>
-#endif
 typedef uintptr_t GUIntptr_t;
 #elif SIZEOF_VOIDP == 8
 typedef GUIntBig GUIntptr_t;
@@ -309,7 +299,7 @@ typedef GUIntBig GUIntptr_t;
 typedef unsigned int  GUIntptr_t;
 #endif
 
-#define CPL_IS_ALIGNED(ptr, quant) (((GUIntptr_t)(ptr) % (quant)) == 0)
+#define CPL_IS_ALIGNED(ptr, quant) ((CPL_REINTERPRET_CAST(GUIntptr_t, CPL_STATIC_CAST(const void*, ptr)) % (quant)) == 0)
 
 #endif
 
@@ -328,19 +318,14 @@ typedef unsigned int  GUIntptr_t;
 #define CPL_FRMT_GUIB    "%" CPL_FRMT_GB_WITHOUT_PREFIX "u"
 
 /*! @cond Doxygen_Suppress */
-/* Workaround VC6 bug */
-#if defined(_MSC_VER) && (_MSC_VER <= 1200)
-#define GUINTBIG_TO_DOUBLE(x) (double)(GIntBig)(x)
-#else
-#define GUINTBIG_TO_DOUBLE(x) (double)(x)
-#endif
+#define GUINTBIG_TO_DOUBLE(x) CPL_STATIC_CAST(double, x)
 /*! @endcond */
 
 /*! @cond Doxygen_Suppress */
 #ifdef COMPAT_WITH_ICC_CONVERSION_CHECK
 #define CPL_INT64_FITS_ON_INT32(x) ((x) >= INT_MIN && (x) <= INT_MAX)
 #else
-#define CPL_INT64_FITS_ON_INT32(x) (((GIntBig)(int)(x)) == (x))
+#define CPL_INT64_FITS_ON_INT32(x) (CPL_STATIC_CAST(GIntBig, CPL_STATIC_CAST(int, x)) == (x))
 #endif
 /*! @endcond */
 
@@ -359,14 +344,29 @@ typedef unsigned int  GUIntptr_t;
 
 #ifndef CPL_DLL
 #if defined(_MSC_VER) && !defined(CPL_DISABLE_DLL)
-#  define CPL_DLL     __declspec(dllexport)
-#else
-#  if defined(USE_GCC_VISIBILITY_FLAG)
-#    define CPL_DLL     __attribute__ ((visibility("default")))
+#  ifdef GDAL_COMPILATION
+#    define CPL_DLL __declspec(dllexport)
 #  else
 #    define CPL_DLL
 #  endif
+#  define CPL_INTERNAL
+#else
+#  if defined(USE_GCC_VISIBILITY_FLAG)
+#    define CPL_DLL     __attribute__ ((visibility("default")))
+#    if !defined(__MINGW32__)
+#        define CPL_INTERNAL __attribute__((visibility("hidden")))
+#    else
+#        define CPL_INTERNAL
+#    endif
+#  else
+#    define CPL_DLL
+#    define CPL_INTERNAL
+#  endif
 #endif
+
+// Marker for unstable API
+#define CPL_UNSTABLE_API CPL_DLL
+
 #endif
 
 /*! @cond Doxygen_Suppress */
@@ -407,52 +407,6 @@ typedef unsigned int  GUIntptr_t;
 #endif
 /*! @endcond*/
 
-/*! @cond Doxygen_Suppress */
-// Define NULL_AS_NULLPTR together with -std=c++11 -Wzero-as-null-pointer-constant with GCC
-// to detect misuses of NULL
-#if defined(NULL_AS_NULLPTR) && HAVE_CXX11
-
-#ifdef __GNUC__
-// We need to include all that bunch of system headers, otherwise
-// as they include <stddef.h> with __need_NULL, this overrides our #define NULL nullptr
-// with #define NULL __null
-#include <locale.h>
-#include <unistd.h>
-#include <sys/types.h>
-#ifdef HAVE_ICONV
-#include <iconv.h>
-#endif
-#ifdef HAVE_MMAP
-#include <sys/mman.h>
-#endif
-#include <signal.h>
-#ifndef _WIN32
-#include <dlfcn.h>
-#include <netdb.h>
-#include <fcntl.h>
-#endif
-
-extern "C++" {
-#include <string>
-#include <cstdio>
-#include <cstdlib>
-#include <cstring>
-#include <cstddef>
-#include <ostream>
-#include <iostream>
-#include <sstream>
-}
-#endif /* __GNUC__ */
-
-#undef NULL
-#define NULL nullptr
-#else /* defined(NULL_AS_NULLPTR) && HAVE_CXX11 */
-#ifndef NULL
-#  define NULL  0
-#endif
-#endif /* defined(NULL_AS_NULLPTR) && HAVE_CXX11 */
-/*! @endcond */
-
 #ifndef MAX
 /** Macro to compute the minimum of 2 values */
 #  define MIN(a,b)      (((a)<(b)) ? (a) : (b))
@@ -608,8 +562,8 @@ static inline char* CPL_afl_friendly_strstr(const char* haystack, const char* ne
 #endif /* defined(AFL_FRIENDLY) && defined(__GNUC__) */
 
 #  if defined(WIN32)
-#    define STRCASECMP(a,b)         (stricmp(a,b))
-#    define STRNCASECMP(a,b,n)      (strnicmp(a,b,n))
+#    define STRCASECMP(a,b)         (_stricmp(a,b))
+#    define STRNCASECMP(a,b,n)      (_strnicmp(a,b,n))
 #  else
 /** Alias for strcasecmp() */
 #    define STRCASECMP(a,b)         (strcasecmp(a,b))
@@ -653,10 +607,17 @@ static inline char* CPL_afl_friendly_strstr(const char* haystack, const char* ne
 #  define CPLIsNan(x) _isnan(x)
 #  define CPLIsInf(x) (!_isnan(x) && !_finite(x))
 #  define CPLIsFinite(x) _finite(x)
-#elif defined(__cplusplus) && defined(__MINGW32__) &&  __GNUC__ == 4 && __GNUC_MINOR__ == 2
-/* Hack for compatibility with ancient i586-mingw32msvc toolchain */
+#elif defined(__GNUC__) && ( __GNUC__ > 4 || ( __GNUC__ == 4 && __GNUC_MINOR__ >= 4 ) )
+/* When including <cmath> in C++11 the isnan() macro is undefined, so that */
+/* std::isnan() can work (#6489). This is a GCC specific workaround for now. */
+#  define CPLIsNan(x)    __builtin_isnan(x)
+#  define CPLIsInf(x)    __builtin_isinf(x)
+#  define CPLIsFinite(x) __builtin_isfinite(x)
+#elif defined(__cplusplus) && defined(HAVE_STD_IS_NAN) && HAVE_STD_IS_NAN
 extern "C++" {
+#ifndef DOXYGEN_SKIP
 #include <cmath>
+#endif
 static inline int CPLIsNan(float f) { return std::isnan(f); }
 static inline int CPLIsNan(double f) { return std::isnan(f); }
 static inline int CPLIsInf(float f) { return std::isinf(f); }
@@ -664,15 +625,9 @@ static inline int CPLIsInf(double f) { return std::isinf(f); }
 static inline int CPLIsFinite(float f) { return std::isfinite(f); }
 static inline int CPLIsFinite(double f) { return std::isfinite(f); }
 }
-#elif defined(__GNUC__) && ( __GNUC__ > 4 || ( __GNUC__ == 4 && __GNUC_MINOR__ >= 4 ) )
-/* When including <cmath> in C++11 the isnan() macro is undefined, so that */
-/* std::isnan() can work (#6489). This is a GCC specific workaround for now. */
-#  define CPLIsNan(x)    __builtin_isnan(x)
-#  define CPLIsInf(x)    __builtin_isinf(x)
-#  define CPLIsFinite(x) __builtin_isfinite(x)
 #else
 /** Return whether a floating-pointer number is NaN */
-#if defined(__cplusplus) && defined(__GNUC__) && defined(__linux) && !defined(__ANDROID__)
+#if defined(__cplusplus) && defined(__GNUC__) && defined(__linux) && !defined(__ANDROID__) && !defined(CPL_SUPRESS_CPLUSPLUS)
 /* so to not get warning about conversion from double to float with */
 /* gcc -Wfloat-conversion when using isnan()/isinf() macros */
 extern "C++" {
@@ -686,7 +641,7 @@ static inline int CPLIsFinite(double f) { return !__isnan(f) && !__isinf(f); }
 #else
 #  define CPLIsNan(x) isnan(x)
 #  if defined(isinf) || defined(__FreeBSD__)
-/** Return whether a floating-pointer number is +/- infinty */
+/** Return whether a floating-pointer number is +/- infinity */
 #    define CPLIsInf(x) isinf(x)
 /** Return whether a floating-pointer number is finite */
 #    define CPLIsFinite(x) (!isnan(x) && !isinf(x))
@@ -724,7 +679,7 @@ static inline int CPLIsFinite(double f) { return !__isnan(f) && !__isinf(f); }
 #endif
 /*! @endcond */
 
-#ifdef __cplusplus
+#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS)
 
 /*! @cond Doxygen_Suppress */
 extern "C++" {
@@ -752,7 +707,7 @@ template<> struct CPLStaticAssert<true>
  *--------------------------------------------------------------------*/
 
 /** Byte-swap a 16bit unsigned integer */
-#define CPL_SWAP16(x) ((GUInt16)( ((GUInt16)(x) << 8) | ((GUInt16)(x) >> 8) ))
+#define CPL_SWAP16(x) CPL_STATIC_CAST(GUInt16, (CPL_STATIC_CAST(GUInt16, x) << 8) | (CPL_STATIC_CAST(GUInt16, x) >> 8) )
 
 #if defined(HAVE_GCC_BSWAP) && (defined(__i386__) || defined(__x86_64__))
 /* Could potentially be extended to other architectures but must be checked */
@@ -760,45 +715,45 @@ template<> struct CPLStaticAssert<true>
 /* GCC (at least 4.6  or above) need that include */
 #include <x86intrin.h>
 /** Byte-swap a 32bit unsigned integer */
-#define CPL_SWAP32(x) ((GUInt32)(__builtin_bswap32((GUInt32)(x))))
+#define CPL_SWAP32(x) CPL_STATIC_CAST(GUInt32, __builtin_bswap32(CPL_STATIC_CAST(GUInt32, x)))
 /** Byte-swap a 64bit unsigned integer */
-#define CPL_SWAP64(x) ((GUInt64)(__builtin_bswap64((GUInt64)(x))))
+#define CPL_SWAP64(x) CPL_STATIC_CAST(GUInt64, __builtin_bswap64(CPL_STATIC_CAST(GUInt64, x)))
 #elif defined(_MSC_VER)
-#define CPL_SWAP32(x) ((GUInt32)(_byteswap_ulong((GUInt32)(x))))
-#define CPL_SWAP64(x) ((GUInt64)(_byteswap_uint64((GUInt64)(x))))
+#define CPL_SWAP32(x) CPL_STATIC_CAST(GUInt32, _byteswap_ulong(CPL_STATIC_CAST(GUInt32, x)))
+#define CPL_SWAP64(x) CPL_STATIC_CAST(GUInt64, _byteswap_uint64(CPL_STATIC_CAST(GUInt64, x)))
 #else
 /** Byte-swap a 32bit unsigned integer */
 #define CPL_SWAP32(x) \
-        ((GUInt32)( \
-            (((GUInt32)(x) & (GUInt32)0x000000ffUL) << 24) | \
-            (((GUInt32)(x) & (GUInt32)0x0000ff00UL) <<  8) | \
-            (((GUInt32)(x) & (GUInt32)0x00ff0000UL) >>  8) | \
-            (((GUInt32)(x) & (GUInt32)0xff000000UL) >> 24) ))
+        CPL_STATIC_CAST(GUInt32, \
+            ((CPL_STATIC_CAST(GUInt32, x) & 0x000000ffU) << 24) | \
+            ((CPL_STATIC_CAST(GUInt32, x) & 0x0000ff00U) <<  8) | \
+            ((CPL_STATIC_CAST(GUInt32, x) & 0x00ff0000U) >>  8) | \
+            ((CPL_STATIC_CAST(GUInt32, x) & 0xff000000U) >> 24) )
 
 /** Byte-swap a 64bit unsigned integer */
 #define CPL_SWAP64(x) \
-            (((GUInt64)(CPL_SWAP32((GUInt32)(x))) << 32) | \
-             (GUInt64)(CPL_SWAP32((GUInt32)((GUInt64)(x) >> 32))))
+            ((CPL_STATIC_CAST(GUInt64, CPL_SWAP32(CPL_STATIC_CAST(GUInt32, x))) << 32) | \
+             (CPL_STATIC_CAST(GUInt64, CPL_SWAP32(CPL_STATIC_CAST(GUInt32, CPL_STATIC_CAST(GUInt64, x) >> 32)))))
 
 #endif
 
 /** Byte-swap a 16 bit pointer */
 #define CPL_SWAP16PTR(x) \
-{                                                                 \
-    GByte       byTemp, *_pabyDataT = (GByte *) (x);              \
+do {                                                              \
+    GByte       byTemp, *_pabyDataT = CPL_REINTERPRET_CAST(GByte*, x);              \
     CPL_STATIC_ASSERT_IF_AVAILABLE(sizeof(*(x)) == 1 || sizeof(*(x)) == 2); \
                                                                   \
     byTemp = _pabyDataT[0];                                       \
     _pabyDataT[0] = _pabyDataT[1];                                \
     _pabyDataT[1] = byTemp;                                       \
-}
+} while(0)
 
 #if defined(MAKE_SANITIZE_HAPPY) || !(defined(__i386__) || defined(__x86_64__) || defined(_M_IX86) || defined(_M_X64))
 
 /** Byte-swap a 32 bit pointer */
 #define CPL_SWAP32PTR(x) \
-{                                                                 \
-    GByte       byTemp, *_pabyDataT = (GByte *) (x);              \
+do {                                                              \
+    GByte       byTemp, *_pabyDataT = CPL_REINTERPRET_CAST(GByte*, x);              \
     CPL_STATIC_ASSERT_IF_AVAILABLE(sizeof(*(x)) == 1 || sizeof(*(x)) == 4);  \
                                                                   \
     byTemp = _pabyDataT[0];                                       \
@@ -807,12 +762,12 @@ template<> struct CPLStaticAssert<true>
     byTemp = _pabyDataT[1];                                       \
     _pabyDataT[1] = _pabyDataT[2];                                \
     _pabyDataT[2] = byTemp;                                       \
-}
+} while(0)
 
 /** Byte-swap a 64 bit pointer */
 #define CPL_SWAP64PTR(x) \
-{                                                                 \
-    GByte       byTemp, *_pabyDataT = (GByte *) (x);              \
+do {                                                              \
+    GByte       byTemp, *_pabyDataT = CPL_REINTERPRET_CAST(GByte*, x);              \
     CPL_STATIC_ASSERT_IF_AVAILABLE(sizeof(*(x)) == 1 || sizeof(*(x)) == 8); \
                                                                   \
     byTemp = _pabyDataT[0];                                       \
@@ -827,31 +782,31 @@ template<> struct CPLStaticAssert<true>
     byTemp = _pabyDataT[3];                                       \
     _pabyDataT[3] = _pabyDataT[4];                                \
     _pabyDataT[4] = byTemp;                                       \
-}
+} while(0)
 
 #else
 
 /** Byte-swap a 32 bit pointer */
 #define CPL_SWAP32PTR(x) \
-{                                                                           \
+do {                                                                        \
     GUInt32 _n32;                                                           \
     void* _lx = x;                                                          \
     memcpy(&_n32, _lx, 4);                                                  \
     CPL_STATIC_ASSERT_IF_AVAILABLE(sizeof(*(x)) == 1 || sizeof(*(x)) == 4); \
     _n32 = CPL_SWAP32(_n32);                                                \
     memcpy(_lx, &_n32, 4);                                                  \
-}
+} while(0)
 
 /** Byte-swap a 64 bit pointer */
 #define CPL_SWAP64PTR(x) \
-{                                                                           \
+do {                                                                        \
     GUInt64 _n64;                                                           \
     void* _lx = x;                                                          \
     memcpy(&_n64, _lx, 8);                                                    \
     CPL_STATIC_ASSERT_IF_AVAILABLE(sizeof(*(x)) == 1 || sizeof(*(x)) == 8); \
     _n64 = CPL_SWAP64(_n64);                                                \
     memcpy(_lx, &_n64, 8);                                                    \
-}
+} while(0)
 
 #endif
 
@@ -895,25 +850,25 @@ template<> struct CPLStaticAssert<true>
 /** Return a Int16 from the 2 bytes ordered in LSB order at address x.
  * @deprecated Use rather CPL_LSBSINT16PTR or CPL_LSBUINT16PTR for explicit
  * signedness. */
-#define CPL_LSBINT16PTR(x)    ((*(GByte*)(x)) | (*(((GByte*)(x))+1) << 8))
+#define CPL_LSBINT16PTR(x)    ((*CPL_REINTERPRET_CAST(const GByte*, x)) | (*((CPL_REINTERPRET_CAST(const GByte*, x))+1) << 8))
 
 /** Return a Int32 from the 4 bytes ordered in LSB order at address x.
  * @deprecated Use rather CPL_LSBSINT32PTR or CPL_LSBUINT32PTR for explicit
  * signedness. */
-#define CPL_LSBINT32PTR(x)    ((*(GByte*)(x)) | (*(((GByte*)(x))+1) << 8) | \
-                              (*(((GByte*)(x))+2) << 16) | (*(((GByte*)(x))+3) << 24))
+#define CPL_LSBINT32PTR(x)    ((*CPL_REINTERPRET_CAST(const GByte*, x)) | (*((CPL_REINTERPRET_CAST(const GByte*, x))+1) << 8) | \
+                              (*((CPL_REINTERPRET_CAST(const GByte*, x))+2) << 16) | (*((CPL_REINTERPRET_CAST(const GByte*, x))+3) << 24))
 
 /** Return a signed Int16 from the 2 bytes ordered in LSB order at address x */
-#define CPL_LSBSINT16PTR(x) ((GInt16) CPL_LSBINT16PTR(x))
+#define CPL_LSBSINT16PTR(x) CPL_STATIC_CAST(GInt16,CPL_LSBINT16PTR(x))
 
 /** Return a unsigned Int16 from the 2 bytes ordered in LSB order at address x */
-#define CPL_LSBUINT16PTR(x) ((GUInt16)CPL_LSBINT16PTR(x))
+#define CPL_LSBUINT16PTR(x) CPL_STATIC_CAST(GUInt16, CPL_LSBINT16PTR(x))
 
 /** Return a signed Int32 from the 4 bytes ordered in LSB order at address x */
-#define CPL_LSBSINT32PTR(x) ((GInt32) CPL_LSBINT32PTR(x))
+#define CPL_LSBSINT32PTR(x) CPL_STATIC_CAST(GInt32, CPL_LSBINT32PTR(x))
 
 /** Return a unsigned Int32 from the 4 bytes ordered in LSB order at address x */
-#define CPL_LSBUINT32PTR(x) ((GUInt32)CPL_LSBINT32PTR(x))
+#define CPL_LSBUINT32PTR(x) CPL_STATIC_CAST(GUInt32, CPL_LSBINT32PTR(x))
 
 /*! @cond Doxygen_Suppress */
 /* Utility macro to explicitly mark intentionally unreferenced parameters. */
@@ -968,27 +923,20 @@ static const char *cvsid_aw() { return( cvsid_aw() ? NULL : cpl_cvsid ); }
 #define CPL_SCAN_FUNC_FORMAT( format_idx, arg_idx )
 #endif
 
-#if defined(_MSC_VER) && _MSC_VER >= 1400 && (defined(GDAL_COMPILATION) || defined(CPL_ENABLE_MSVC_ANNOTATIONS))
+#if defined(_MSC_VER) && (defined(GDAL_COMPILATION) || defined(CPL_ENABLE_MSVC_ANNOTATIONS))
 #include <sal.h>
-# if _MSC_VER > 1400
 /** Macro into which to wrap the format argument of a printf-like function.
  * Only used if ANALYZE=1 is specified to nmake */
 #  define CPL_FORMAT_STRING(arg) _Printf_format_string_ arg
 /** Macro into which to wrap the format argument of a sscanf-like function.
  * Only used if ANALYZE=1 is specified to nmake */
 #  define CPL_SCANF_FORMAT_STRING(arg) _Scanf_format_string_ arg
-# else
-/** Macro into which to wrap the format argument of a printf-like function */
-#  define CPL_FORMAT_STRING(arg) __format_string arg
-/** Macro into which to wrap the format argument of a sscanf-like function. */
-#  define CPL_SCANF_FORMAT_STRING(arg) arg
-# endif
 #else
 /** Macro into which to wrap the format argument of a printf-like function */
 # define CPL_FORMAT_STRING(arg) arg
 /** Macro into which to wrap the format argument of a sscanf-like function. */
 # define CPL_SCANF_FORMAT_STRING(arg) arg
-#endif /* defined(_MSC_VER) && _MSC_VER >= 1400 && defined(GDAL_COMPILATION) */
+#endif /* defined(_MSC_VER) && defined(GDAL_COMPILATION) */
 
 #if defined(__GNUC__) && __GNUC__ >= 4 && !defined(DOXYGEN_SKIP)
 /** Qualifier to warn when the return value of a function is not used */
@@ -1039,31 +987,18 @@ static const char *cvsid_aw() { return( cvsid_aw() ? NULL : cpl_cvsid ); }
 #define CPL_RESTRICT
 #endif
 
-#ifdef __cplusplus
-
-#if HAVE_CXX11 || _MSC_VER >= 1500
+#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS)
 
 /** To be used in public headers only. For non-public headers or .cpp files,
  * use override directly. */
 #  define CPL_OVERRIDE override
 
-#else
-
-/** To be used in public headers only. For non-public headers or .cpp files,
- * use override directly. */
-#  define CPL_OVERRIDE
-
-/* For GDAL source compilation only, ignore override if non C++11 compiler */
-#ifdef GDAL_COMPILATION
-#  define override
-#endif
-
-#endif /* HAVE_CXX11 || _MSC_VER >= 1500 */
-
-#if HAVE_CXX11
 /** C++11 final qualifier */
 #  define CPL_FINAL final
 
+/** Mark that a class is explicitly recognized as non-final */
+#  define CPL_NON_FINAL
+
 /** Helper to remove the copy and assignment constructors so that the compiler
    will not generate the default versions.
 
@@ -1072,19 +1007,6 @@ static const char *cvsid_aw() { return( cvsid_aw() ? NULL : cpl_cvsid ); }
 #  define CPL_DISALLOW_COPY_ASSIGN(ClassName) \
     ClassName( const ClassName & ) = delete; \
     ClassName &operator=( const ClassName & ) = delete;
-#else
-/** C++11 final qualifier */
-#  define CPL_FINAL
-
-/** Helper to remove the copy and assignment constructors so that the compiler
-   will not generate the default versions.
-
-   Must be placed in the private section of a class and should be at the end.
-*/
-#  define CPL_DISALLOW_COPY_ASSIGN(ClassName) \
-    ClassName( const ClassName & ); \
-    ClassName &operator=( const ClassName & );
-#endif  /* HAVE_CXX11 */
 
 #endif /* __cplusplus */
 
@@ -1128,14 +1050,14 @@ CPL_C_END
 /*! @endcond */
 #endif
 
-#ifdef __cplusplus
+#if defined(__cplusplus)
 /** Returns the size of C style arrays. */
 #define CPL_ARRAYSIZE(array) \
   ((sizeof(array) / sizeof(*(array))) / \
   static_cast<size_t>(!(sizeof(array) % sizeof(*(array)))))
 
 extern "C++" {
-template<class T> static void CPL_IGNORE_RET_VAL(T) {}
+template<class T> static void CPL_IGNORE_RET_VAL(const T&) {}
 inline static bool CPL_TO_BOOL(int x) { return x != 0; }
 } /* extern "C++" */
 
@@ -1149,7 +1071,7 @@ inline static bool CPL_TO_BOOL(int x) { return x != 0; }
 #define HAVE_GCC_SYSTEM_HEADER
 #endif
 
-#if ((defined(__clang__) && (__clang_major__ > 3 || (__clang_major__ == 3 && __clang_minor__ >=7))) || __GNUC__ >= 7) && HAVE_CXX11
+#if ((defined(__clang__) && (__clang_major__ > 3 || (__clang_major__ == 3 && __clang_minor__ >=7))) || __GNUC__ >= 7)
 /** Macro for fallthrough in a switch case construct */
 #  define CPL_FALLTHROUGH [[clang::fallthrough]];
 #else
@@ -1162,7 +1084,7 @@ inline static bool CPL_TO_BOOL(int x) { return x != 0; }
 // a integer is assigned to a bool
 // WARNING: use only at compilation time, since it is know to not work
 //  at runtime for unknown reasons (crash in MongoDB driver for example)
-#if defined(__cplusplus) && defined(DEBUG_BOOL) && !defined(DO_NOT_USE_DEBUG_BOOL)
+#if defined(__cplusplus) && defined(DEBUG_BOOL) && !defined(DO_NOT_USE_DEBUG_BOOL) && !defined(CPL_SUPRESS_CPLUSPLUS)
 extern "C++" {
 class MSVCPedanticBool
 {
@@ -1186,12 +1108,15 @@ class MSVCPedanticBool
 
         bool operator== (const bool& other) const { return b == other; }
         bool operator!= (const bool& other) const { return b != other; }
+        bool operator< (const bool& other) const { return b < other; }
         bool operator== (const MSVCPedanticBool& other) const { return b == other.b; }
         bool operator!= (const MSVCPedanticBool& other) const { return b != other.b; }
+        bool operator< (const MSVCPedanticBool& other) const { return b < other.b; }
 
         bool operator! () const { return !b; }
         operator bool() const { return b; }
         operator int() const { return b; }
+        operator GIntBig() const { return b; }
 };
 
 inline bool operator== (const bool& one, const MSVCPedanticBool& other) { return one == other.b; }
@@ -1207,6 +1132,15 @@ inline bool operator!= (const bool& one, const MSVCPedanticBool& other) { return
 #include <sstream>
 #include <fstream>
 #include <algorithm>
+#include <functional>
+#include <memory>
+#include <queue>
+#include <mutex>
+#include <unordered_map>
+#include <thread>
+#include <unordered_set>
+#include <complex>
+#include <iomanip>
 
 } /* extern C++ */
 
@@ -1239,6 +1173,47 @@ inline bool operator!= (const bool& one, const MSVCPedanticBool& other) { return
 #define VOLATILE_BOOL volatile bool
 
 #endif /* defined(__cplusplus) && defined(DEBUG_BOOL) */
+
+#if __clang_major__ >= 4 || (__clang_major__ == 3 && __clang_minor__ >= 8)
+#define CPL_NOSANITIZE_UNSIGNED_INT_OVERFLOW __attribute__((no_sanitize("unsigned-integer-overflow")))
+#else
+#define CPL_NOSANITIZE_UNSIGNED_INT_OVERFLOW
+#endif
+
+#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS) && defined(GDAL_COMPILATION)
+extern "C++" {
+template<class C, class A, class B>
+CPL_NOSANITIZE_UNSIGNED_INT_OVERFLOW
+inline C CPLUnsanitizedAdd(A a, B b)
+{
+    return a + b;
+}
+}
+#endif
+
 /*! @endcond */
 
+/*! @cond Doxygen_Suppress */
+#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS)
+#define CPL_NULLPTR nullptr
+#else
+#define CPL_NULLPTR NULL
+#endif
+/*! @endcond */
+
+/* This typedef is for C functions that take char** as argument, but */
+/* with the semantics of a const list. In C, char** is not implicitly cast to */
+/* const char* const*, contrary to C++. So when seen for C++, it is OK */
+/* to expose the prototypes as const char* const*, but for C we keep the */
+/* historical definition to avoid warnings. */
+#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS) && !defined(DOXYGEN_SKIP)
+/** Type of a constant null-terminated list of nul terminated strings.
+ * Seen as char** from C and const char* const* from C++ */
+typedef const char* const* CSLConstList;
+#else
+/** Type of a constant null-terminated list of nul terminated strings.
+ * Seen as char** from C and const char* const* from C++ */
+typedef char** CSLConstList;
+#endif
+
 #endif /* ndef CPL_BASE_H_INCLUDED */
